# Allgemeines

![Wafflebot3](https://emanual.robotis.com/assets/images/platform/turtlebot3/hardware_setup/turtlebot3_models.png)

In dieser Zusammenfassung geht es um den sogenannten Turtlebot3 Wafflebot (der rechte Roboter im obigen Bild) und wie
dieser durch Skripte in Verbindung mit Nodes und Topics angesteuert wird.  
Zuerst geht es darum, das Grundprinzip hinter einer Node und einem Topic zu verstehen.

## Node

Eine Node-basierte Software, beispielsweise die, die den Wafflebot fahren lässt, wird in verschiedene Einzelteile
aufgeteilt. Hierfür wird jeder vergebenen Aufgabe genau eine Node zugeordnet. Somit sind Nodes im Prinzip die Bauteile
eines ROS-Skriptes. Eine Beispielaufgabe hierfür wäre das Auslesen der Sensordaten des Lasers, der sich auf dem
Wafflebot befindet. Um nun fahren zu können, muss der Roboter die Motoren der einzelnen Reifen ansteuern. Auch das sind
Aufgaben, die je eine Node übernehmen kann. Beispiel für das Erstellen einer neuen Node anhand der im Git enthaltenen
Turtlebot3 Node:

```python
import rclpy  # ROS client library
from rclpy.node import Node

class Tb3(Node):
    def __init__(self):
        super().__init__('tb3')
    
    def main(args=None):
        rclpy.init(args=args)  # registriert die Node im aktuellen ROS-Environment
        tb3 = Tb3()  # erstellt ein neues Objekt der Tb3-Klasse namens tb3
        try:
            rclpy.spin(tb3)  # führt die tb3 Node aus
            # Blockiert solange, bis die rclpy.spin(tb3) Methode nicht mehr ausgeführt wird
        except KeyboardInterrupt:
            pass
        
        tb3.destroy_node()  # löscht die aktuelle Node
        rclpy.shutdown()  # beendet das aktuell laufende Node-Programm
``` 


Die einzelnen Code Snippets lassen sich folgendermaßen erklären:  

* `rclpy.init(args=args)`: Die neu zu erstellende Node wird im aktuellen ROS-Environment registriert
* `tb3 = Tb3()`: Die Node mit dem Namen `tb3` wird hiermit als neues Objekt der Klasse `Tb3()` erstellt
* `rclpy.spin(tb3)`: Führt die Node solange aus, bis sie durch das `except`-Statement -und somit durch einen
KeyboardInterrupt- gestoppt wird

## Topic

Was Nodes sind, wurde bereits erklärt und ein Topic benötigt diese Nodes.
Allgemein ist zu sagen, dass ein Topic eine Publisher-Node und eine Subscriber-Node hat. Der Publisher veröffentlicht 
(engl. "to publish") eine Nachricht im Topic und der Subscriber abonniert (engl. "subscribe") das Topic und erhält somit
alle neuen, im Topic abgelegten Nachricht.

Um die Funktionalität eines Topics zu erklären,
habe ich folgendes Bild erstellt:

![Topic und Nodes](source/Topic_und_Nodes.jpeg)

Node1 ist Publisher (rot) des Topics und Node2 ist Subscriber (rot) des Topics. Node1 sendet eine Nachricht, bspw. einen
Sensorwert des Lasers vom Wafflebot (gelb), an das Topic. Da Node2 der Subscriber des Topics ist, erhält diese den in 
das Topic veröffentlichte Sensorwert.
Somit erschließt sich, dass die Nodes-zu-Topic Beziehung im Prinzip eine Publisher-Subscriber Beziehung ist.

### Code-Beispiel Publisher
Das Code-Beispiel wird wieder aus dem Turtlebot3 Repo genommen:
```python
from geometry_msgs.msg import Twist

def __init__(self):
    super().__init__('tb3')
    
    self.cmd_vel_pub = self.create_publisher(
            Twist,      # Art der Nachricht
            'cmd_vel',  # Name des Topics
            1)          # Queue Tiefe

def node_publish(self, msg):
    self.cmd_vel_pub.publish(msg)
```
Hierfür wird beim Initialisieren der neuen Tb3-Instanz, vorher `tb3` genannt, ein Publisher mit initialisiert.
* `self.create_publisher()`: Erstellt eine neue Publisher-Node mit drei Parametern:
  * `Twist` ist die Art der Nachricht, welche in diesem Beispiel von `geometry_msgs.msg` importiert wird; andere
Beispiele wären Integer, String, Float, etc.
  * `'cmd_vel'`: Name des Topics
  * `1`: Die Tiefe der Queue Size des Topics

* `node_publish(self, msg)` ist eine extra Methode für das Publishen einer Nachricht, welche eine Nachricht `msg` als
Übergabeparameter benötigt; in der Methode wird dann die Nachricht veröffentlicht:
  * `self.cmd_vel_pub.publish(msg)`: Hier wird die im Konstruktor erstellte `self.cmd_vel_pub` Publisher-Node genutzt,
um mit der Methode `.publish()` die vorgenannte Nachricht zu veröffentlichen
  
### Code-Beispiel Subscriber
Auch für dieses Code-Beispiel wird der Code aus dem Turtlebot3 Repo genommen:
```python
from sensor_msgs.msg import LaserScan

def __init__(self):
    super().__init__('tb3')
    
    self.scan_sub = self.create_subscription(
            LaserScan,
            'scan',
            self.scan_callback,  # Funktion, die bei Ankunft einer Nachricht aufgerufen wird
            qos_profile_sensor_data)  # Erlaubt Paketverlust

def scan_callback(self, msg):
    doSomething
```
Hierfür wird, wie bereits beim Publisher, beim Initialisieren der neuen Tb3-Instanz, vorher `tb3` genannt, ein 
Subscriber mit initialisiert.
* `self.create_subscription()`: Erstellt eine neue Publisher-Node mit vier Parametern:
  * `LaserScan` ist wieder die Art der Nachricht
  * `'scan'`: Der Name des zu abonnierenden Topics
  * `self.scan_callback` ist die Methode, die bei jeder Ankunft einer neuen Nachricht aufgerufen wird
  * `qos_profile_sensor_data`: Erlaubt möglichen Paketverlust beim Publish-Subscribe Vorgang

* `scan_callback(self, msg)` ist die aufgerufene Methode. Hier gibt es als Übergabeparameter wieder die übergebene
Nachricht `msg`. In dieser Methode wird alles ausgeführt, was zum Fahren des Roboters gecoded wird. `doSomething` ist
hierbei lediglich ein Platzhalter.